# Copyright 2025 Snowflake Inc.
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import pandas as pd

from snowflake.snowflake_data_validation.configuration.model.table_configuration import (
    TableConfiguration,
)
from snowflake.snowflake_data_validation.connector.connector_base import (
    ConnectorBase,
)
from snowflake.snowflake_data_validation.query.query_generator_base import (
    QueryGeneratorBase,
)
from snowflake.snowflake_data_validation.utils.constants import (
    COLUMN_METADATA_QUERIES_FILENAME,
    TABLE_METADATA_QUERIES_FILENAME,
)
from snowflake.snowflake_data_validation.utils.context import Context
from snowflake.snowflake_data_validation.utils.helpers import write_to_file


class ScriptWriterBase:

    """Base class for printing database queries to files.

    This class provides the same interface as MetadataExtractorBase but instead of
    returning pandas DataFrames, it writes the generated SQL queries to files.

    All methods have concrete implementations since the logic is identical across
    all database platforms - they create TableConfiguration objects and delegate
    to the platform-specific query generators.
    """

    def __init__(
        self,
        connector: ConnectorBase,
        query_generator: QueryGeneratorBase,
        report_path: str = "",
    ):
        """Initialize the script printer with a database connector and query generator.

        Args:
            connector: Database connector instance for the specific database type.
            query_generator: Query generator instance for generating SQL queries.
            report_path: Path where the script files should be written.

        """
        self.connector = connector
        self.query_generator = query_generator
        self.report_path = report_path or "."

    def _get_filename(self, filename_template: str, context: Context) -> str:
        """Get the filename with timestamp from context.

        Args:
            filename_template (str): The filename template with {timestamp} placeholder.
            context (Context): The execution context containing run_start_time.

        Returns:
            str: Full file path with timestamp.

        """
        timestamp = context.run_start_time
        filename = filename_template.format(timestamp=timestamp)
        return os.path.join(self.report_path, filename)

    def print_table_metadata_query(
        self,
        table_context: TableConfiguration,
        context: Context,
    ) -> None:
        """Write the table-level metadata query for the specified table to file.

        Args:
            table_context (TableConfiguration): The table configuration containing all necessary parameters.
            context (Context): The execution context containing relevant configuration and runtime information.

        """
        query = self.query_generator.generate_schema_query(
            table_context=table_context,
            context=context,
        )
        file_path = self._get_filename(TABLE_METADATA_QUERIES_FILENAME, context)
        write_to_file(file_path, query)

    def print_column_metadata_query(
        self,
        table_context: TableConfiguration,
        metrics_templates: pd.DataFrame,
        context: Context,
    ) -> None:
        """Write the column-level metadata query for the specified columns in a table to file.

        Args:
            table_context (TableConfiguration): The table configuration containing all necessary parameters.
            metrics_templates (pd.DataFrame): DataFrame containing metrics templates to be applied.
            context (Context): The execution context containing relevant configuration and runtime information.

        """
        query = self.query_generator.generate_metrics_query(
            table_context=table_context,
            metrics_templates=metrics_templates,
            context=context,
            connector=self.connector,
        )
        file_path = self._get_filename(COLUMN_METADATA_QUERIES_FILENAME, context)
        write_to_file(file_path, query)
